/*************************************************************************************
 * Description:	TFT_SSD2119 TFT LCD 320x240 Driver module
 *				Written for a ATxMEGA128A1U EBI bus use.
 *				The EBI bus must be run with 2 wait states on the SRAM port,
 *				otherwise the display will not init properly and may have 
 *				junk on the screen when it does init properly. This was 
 *				confirmed with a CPU running at 176MHz Clkper/4 and the EBI bus
 *				clocked at 88MHz Clkper/2 rate.
 *
 *				The EBI bus is run in 2-port LPC SRAM mode with a single
 *				ALE1 line for /C-D selection from Addr bit 0. Port-H and
 *				Port-J are the two I/O ports used on ATxmega128A1U part.
 *				This code should also work with any of the ATmega parts that
 *				support a EBI (XMEM) bus, but that has not been tried by me.
 *
 *---------------------------------------------------------------------------------*
 * Written by:	JWM Engineering Group - Copyright (c) 2017, ALl rights reserved.   *
 *				NOTE: This code may be used by anyone, for any purpose, as long    *
 *				as the copyright notice is included somewhere in your final code.  *
 *---------------------------------------------------------------------------------*
 *
 * Date:		02/25/2017 - JW Mulchin - Owner - Started code .....
 * Date:		02/26/2017 - JW Mulchin - converted over to uGUI code
 * Date:		03/02/2017 - JW Mulchin - Finalized code for release use in uGUI
 *
 ************************************************************************************/
#include "global.h"																// contains mostly common AVR include files

#include "TFT_SSD2119.h"
#include "ugui_config.h"														// need some config info here

// Define our EBI Graphics Controller Port start address
#define CS0_START				INTERNAL_SRAM_END + 1							// starting addr of EBI CS0 chip select

BYTE volatile *CMDptr = (void *)CS0_START;										// 8-bit pointer to SRAM Cmnd port of TFT
BYTE volatile *DATptr = (void *)(CS0_START+1);									// 8-bit pointer to SRAM Data port of TFT

#ifdef USE_ATMEL_AVR		/************  If using an Atmel CPU, Move these INIT values into PROGRAM space  **************/
static WORD const LCDIni[] PROGMEM =
#else
static WORD const LCDIni[] =
#endif
						{														// all page #s refer to SSD2119 datasheet
						SSD2119_REG_VCOM_OTP_1,        0x0006,					// VCOM OTP Page 55-56
						SSD2119_REG_OSC_START,         0x0001,					// start Oscillator Page 36
						SSD2119_REG_SLEEP_MODE_1,      0x0000,					// Sleep mode Page 49
						SSD2119_REG_OUTPUT_CTRL,       0x72EF,					// Driver Output Control Page 36-39
						SSD2119_REG_LCD_DRIVE_AC_CTRL, 0x0600,					// LCD Driving Waveform Control Page 40-42
						SSD2119_REG_PWR_CTRL_1,        0x6A38,					// Power Control 1 Page 43-44
						SSD2119_REG_ENTRY_MODE,        0x6870,					// Entry Mode Page 50-52
						SSD2119_REG_GATE_SCAN_START,   0x0000,					// Gate Scan Position Page 49
						SSD2119_REG_FRAME_CYCLE_CTRL,  0x5308,					// Frame Cycle Control Page 45
						SSD2119_REG_PWR_CTRL_2,        0x0003,					// Power Control 2 Page 47
						SSD2119_REG_PWR_CTRL_3,        0x000A,					// Power Control 3 Page 48
						SSD2119_REG_PWR_CTRL_4,        0x2E00,					// Power Control 4 Page 48
						SSD2119_REG_PWR_CTRL_5,        0x00BE,					// Power Control 5 Page 53
						SSD2119_REG_FRAME_FREQ,        0x8000,					// Frame Frequency Control Page 53
						SSD2119_REG_ANALOG_SET,        0x3800,					// Analog setting Page 54
						SSD2119_REG_V_RAM_POS,         0xEF00,					// Vertical RAM address position Page 57
						SSD2119_REG_H_RAM_START,       0x0000,					// Horizontal RAM address position	Page 57
						SSD2119_REG_H_RAM_END,         0x013F,					// Horizontal RAM address position Page 57
						SSD2119_REG_X_RAM_ADDR,        0x0000,					// Ram Address Set Page 58
						SSD2119_REG_Y_RAM_ADDR,        0x0000,					// Ram Address Set Page 58
						SSD2119_REG_SLEEP_MODE_2,      0x08D9,					// Sleep mode Page 49

						SSD2119_REG_GAMMA_CTRL_1,      0x0000,					// Gamma Control (R30h to R3Bh)
						SSD2119_REG_GAMMA_CTRL_2,      0x0104,					// Page 56
						SSD2119_REG_GAMMA_CTRL_3,      0x0100,
						SSD2119_REG_GAMMA_CTRL_4,      0x0305,
						SSD2119_REG_GAMMA_CTRL_5,      0x0505,
						SSD2119_REG_GAMMA_CTRL_6,      0x0305,
						SSD2119_REG_GAMMA_CTRL_7,      0x0707,
						SSD2119_REG_GAMMA_CTRL_8,      0x0300,
						SSD2119_REG_GAMMA_CTRL_9,      0x1200,
						SSD2119_REG_GAMMA_CTRL_10,     0x0800,

						SSD2119_REG_DISPLAY_CTRL,      0x0033,					// Display Control	Page 45
						SSD2119_INIT_TERMINATOR									// a single terminator word to stop Init loop
						};

/*************************************************
	LCD delay routine
*************************************************/

void LCDdelay(uint32_t t)
{
	unsigned char t1;

	while(t--)
	for(t1 = 11;  t1 > 0;  t1--)
		{
		asm("nop");
		}
}

/***********************************************************
	Initialize display
***********************************************************/

void TFTInit(void)
{
#ifdef USE_ATMEL_AVR
	WORD	ptr = 0;
	WORD	indx = 0;										// index for array use

	PORTH.OUT &= ~TFT_RST;									// set hardware reset LO
	LCDdelay(35000);
	PORTH.OUT = TFT_RST;									// set hardware reset HI

	while ((ptr = pgm_read_word(&LCDIni[indx])) < SSD2119_INIT_TERMINATOR)
		{
		LCDCmd(pgm_read_word(&LCDIni[indx++]));				// fetch TFT command value
		LCDData(pgm_read_word(&LCDIni[indx++]));			// fetch TFT data value
		}
	LCDClear(0, NR_ROWS);									// clear data memory to black screen
#else
	const WORD *ptr;

	PORTH.OUT &= ~TFT_RST;									// set hardware reset LO
	LCDdelay(35000);
	PORTH.OUT = TFT_RST;									// set hardware reset HI

	ptr = &LCDIni[0];										// TFT Init table
	while (*ptr < 0xFFFF)
	{
		LCDCmd(*(ptr++));									// fetch TFT command value
		LCDData(*(ptr++));									// fetch TFT data value
	}
	LCDClear(0, NR_ROWS);									// clear data memory
#endif
}

/***********************************************************
	Write command to LCD controller
***********************************************************/

void LCDCmd(WORD arg)
{
	*CMDptr = arg >> 8;										// send HI byte first
	*CMDptr = arg & 0xFF;									// send LO byte second
}

/***********************************************************
	Write data to LCD controller
***********************************************************/

void LCDData(WORD arg)
{
	*DATptr = arg >> 8;										// send HI byte first
	*DATptr = arg & 0xFF;									// send LO byte second
}

/***********************************************************
	Clear the display data RAM
	row = starting row
	hgt = number of rows
***********************************************************/

void LCDClear(WORD row, WORD hgt)
{
	WORD a, b;

	LCDSetGCur(row, 0);
	for (a = 0; a < hgt; a++)								// each row
		{
		for (b = 0; b < NR_COLS; b++)						// all columns
			LCDData(0x0000);								// clears the screen to BLACK
		}
}

/***********************************************************
	Write cursor address to controller
***********************************************************/

void LCDSetGCur(WORD row, WORD col)
{
	if (col > LASTCOL || row > LASTROW)
		return;

	LCDCmd(SSD2119_REG_X_RAM_ADDR);							// GRAM X address set
	LCDData(col);
	LCDCmd(SSD2119_REG_Y_RAM_ADDR);							// GRAM Y address set
	LCDData(row);
	LCDCmd(SSD2119_REG_RAM_DATA);							// access to display RAM
}

/***************************************************************
	uGUI pset function
	Set a pixel on the SSD2119 TFT Display
***************************************************************/
void ssd2119_pset ( UG_S16 x , UG_S16 y , UG_COLOR c )
{
   LCDCmd(SSD2119_REG_X_RAM_ADDR);							// Set Column Address (Reg 4Eh in SSD2119)
   LCDData(x);												// send Upper then Lower byte
   
   LCDCmd(SSD2119_REG_Y_RAM_ADDR);							// Set Row Address	(Reg 4F in SSD2119)
   LCDData(y);												// send Upper then Lower byte
   
   LCDCmd(SSD2119_REG_RAM_DATA);							// Write RAM command (Reg 22h in SSD2119)

   LCDData(c);												// send HI then LO byte of color
}
